(* ::Package:: *)

(*:Package Version: 1.01 *)

(*:Version History: 3.02 (2016.09.20 - 2016.10.20)NDSolveHill.m
                    1.01 (2016.10.21 -     .  .  )Heun *)

(*:Copyright: Copyright 2016.10-2016.10, Souichiro Ikebe *)

(*:Author: Souichiro Ikebe *)

(*:Name: SpecialFunctions`NDSolveHeun` *)

(*:Title: NDSolve, Heun, ConfluentHeun *)

(*:Keywords: HeunHL, HeunHLPrime, HeunHS, HeunHSPrime, GeneralizedHeun, 
             GeneralizedHeunPrime, ConfluentHeun, ConfluentHeunPrime, 
             DoublyConfluentHeun, DoublyConfluentHeunPrime, BiConfluentHeun, 
             BiConfluentHeunPrime, TriConfluentHeun, TriConfluentHeunPrime *)

(*:Sources:
  H.Volkmer, G.Wolf,(edited: F.W.J.Olver, D.W.Lozier, R.F.Boisvert, C.W.Clark),
  "NIST Handbook of mathematical functions",
  National institute of standards and technology & 
  Cambridge university press, (2010),pp .709-721.
*)

(*:Summary:
\:3000\:3000This package will add a method to solved in the complex domain 
\:3000\:3000the solution of Heun differential equations using the NDSolve, 
\:3000\:3000a feature built into the Mathematica.
   This package allows the calculation of Heun functions and Confluent 
   Heun functions.
*)

(**********************************************************************************)





BeginPackage["SpecialFunctions`NDSolveHeun`"]





HeunHL::usage =
"HeunHL[a,q,{alp,bet,gam,del},z] gives Local Heun functions of first kind."

HeunHLPrime::usage =
"HeunHLPrime[a,q,{alp,bet,gam,del},z] gives derivative of Local Heun functions of 
 first kind at z."

HeunHS::usage =
"HeunHS[a,q,{alp,bet,gam,del},z] gives Local Heun functions of second kind."

HeunHSPrime::usage =
"HeunHSPrime[a,q,{alp,bet,gam,del},z] gives derivative of Local Heun functions of 
 second kind at z."

GeneralizedHeun::usage =
"GeneralizedHeun[a,q,{alp,bet,gam,del},z,{w0,w1}] gives solutions of the Heun 
 differential equation : y''[z]+(gam/z+del/(z-1)+eps/(z-a))*y'[z]+
 (alp*bet*z-q)/(z*(z-1)*(z-a))*y[z]=0; 
 with eps=alp+bet-gam-del+1; y[0]=w0; y'[0]=w1."

GeneralizedHeunPrime::usage =
"GeneralizedHeunPrime[a,q,{alp,bet,gam,del},z,{w0,w1}] gives derivative of the 
 Heun function : GeneralizedHeun[a,q,{alp,bet,gam,del},z,{w0,w1}] at z."

ConfluentHeun::usage =
"ConfluentHeun[q,{alp,gam,del,eps},z,{w0,w1}] gives solutions of the Confluent 
 Heun differential equation : y''[z]+(gam/z+del/(z-1)+eps)*y'[z]+
 (alp*z-q)/(z*(z-1))*y[z]=0; 
 with y[1/2]=w0; y'[1/2]=w1."

ConfluentHeunPrime::usage =
"ConfluentHeunPrime[q,{alp,gam,del,eps},z,{w0,w1}] gives derivative of the 
 Confluent Heun function : ConfluentHeun[q,{alp,gam,del,eps},z,{w0,w1}] at z."

DoublyConfluentHeun::usage =
"DoublyConfluentHeun[q,{alp,gam,del},z,{w0,w1}] gives solutions of the Doubly-
 confluent Heun differential equation : y''[z]+(gam/z+del/(z^2)+1)*y'[z]+
 (alp*z-q)/(z^2)*y[z]=0; 
 with y[1]=w0; y'[1]=w1."

DoublyConfluentHeunPrime::usage =
"DoublyConfluentHeunPrime[q,{alp,gam,del},z,{w0,w1}] gives derivative of the 
 Doubly-confluent Heun function : DoublyConfluentHeun[q,{alp,gam,del},z,{w0,w1}] 
 at z."

BiConfluentHeun::usage =
"BiConfluentHeun[q,{alp,gam,del},z,{w0,w1}] gives solutions of the Biconfluent 
 Heun differential equation : y''[z]+(gam/z+del+z)*y'[z]+(alp*z-q)/z*y[z]=0; 
 with y[1]=w0; y'[1]=w1."

BiConfluentHeunPrime::usage =
"BiConfluentHeunPrime[q,{alp,gam,del},z,{w0,w1}] gives derivative of the 
 Biconfluent Heun function : BiConfluentHeun[q,{alp,gam,del},z,{w0,w1}] at z."

TriConfluentHeun::usage =
"TriConfluentHeun[q,{alp,gam},z,{w0,w1}] gives solutions of the Triconfluent 
 Heun differential equation : y''[z]+z*(gam+z)*y'[z]+(alp*z-q)*y[z]=0; 
 with y[0]=w0; y'[0]=w1."

TriConfluentHeunPrime::usage =
"TriConfluentHeunPrime[q,{alp,gam},z,{w0,w1}] gives derivative of the 
 Triconfluent Heun function : TriConfluentHeun[q,{alp,gam},z,{w0,w1}] at z."





Begin["SpecialFunctions`NDSolveHeun`Private`"]





(**********************************************************************************)
(*                              Local Heun functions                              *)
(**********************************************************************************)

(**************************** Common codes *****************************)

(* Test function of matching integers *)

intQHL[gg_]:=(Chop[Im[gg]]===0 && Chop[gg-Floor[gg+1/2]]===0)

(* Polynomials used in recurrence relations *)

RecurP[n_,{ag_,bg_}]:=(n+ag-1)*(n+bg-1);

RecurQ[n_,{a_,ag_,bg_,gg_,dg_}]:=n*((n+gg-1)*(1+a)+a*dg+ag+bg-gg-dg+1);

RecurR[n_,{a_,gg_}]:=a*(n+1)*(n+gg);

(* Test function of matching branch *)

branchRightQ[a_,z_]:=
  If[Im[a]>0,
    Im[z]>=0 && 
    If[Re[a]>0,Im[z]<=Im[a]/Re[a]*Re[z],
      If[Re[a]<0,
        Im[z]>=Im[a]/Re[a]*Re[z],Re[z]>=0]],
    Im[z]<=0 && 
    If[Re[a]>0,Im[z]>=Im[a]/Re[a]*Re[z],
      If[Re[a]<0,
        Im[z]<=Im[a]/Re[a]*Re[z],Re[z]>=0]]]/;
  (Abs[Im[a]]>0)

branchLeftQ[a_,z_]:=
  If[Im[a]>0,
    Im[z]>=0 && 
    If[Re[a]>0,Im[z]>Im[a]/Re[a]*Re[z],
      If[Re[a]<0,
        Im[z]<Im[a]/Re[a]*Re[z],Re[z]<0]],
    Im[z]<=0 && 
    If[Re[a]>0,Im[z]<Im[a]/Re[a]*Re[z],
      If[Re[a]<0,
        Im[z]>Im[a]/Re[a]*Re[z],Re[z]<0]]]/;
  (Abs[Im[a]]>0)

(************************ HeunHL & HeunHLPrime *************************)

(* Local Heun functions of the 1st kind *)

HeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHL[a,q,{ag,bg,gg,dg},z],
    realAHeunHL[a,q,{ag,bg,gg,dg},z]]/;
  (Not[intQHL[gg] && Re[gg]<1] && Precision[z]<Infinity)

HeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHLPrime[a,q,{ag,bg,gg,dg},z],
    realAHeunHLPrime[a,q,{ag,bg,gg,dg},z]]/;
  (Not[intQHL[gg] && Re[gg]<1] && Precision[z]<Infinity)

realAHeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHL[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHL[a,q,{ag,bg,gg,dg},z,-I]];

realAHeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,-I]];

compAHeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],ndsHeunHL[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]<0,ndsHeunHL[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHL[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
    Which[branchRightQ[a,z],ndsHeunHL[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]>0,ndsHeunHL[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHL[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

compAHeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
If[Im[a]>0,
  Which[branchRightQ[a,z],ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
    Im[z]<0,ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
    branchLeftQ[a,z],ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
  Which[branchRightQ[a,z],ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
    Im[z]>0,ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
    branchLeftQ[a,z],ndsHeunHLPrime[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

ndsHeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHL[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHLPrime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   f[Abs[z-shift]]/.sol[[1]]];

ndsHeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHL[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHLPrime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of Hl(z) *)

initHeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250,prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefHL[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefHLA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefHLB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefHL[a_,q_,{ag_,bg_,gg_,dg_},0]:=1;

coefHL[a_,q_,{ag_,bg_,gg_,dg_},1]:=q/(a*gg);

coefHL[a_,q_,{ag_,bg_,gg_,dg_},n_/;(IntegerQ[n] && n>1)]:=
  coefHL[a,q,{ag,bg,gg,dg},n]=
  ((RecurQ[n-1,{a,ag,bg,gg,dg}]+q)*coefHL[a,q,{ag,bg,gg,dg},n-1]-
   RecurP[n-1,{ag,bg}]*coefHL[a,q,{ag,bg,gg,dg},n-2])/RecurR[n-1,{a,gg}];

padeMatrixHLA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefHL[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixHLB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefHL[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefHLA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHLA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixHLA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefHLB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefHLB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHLB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixHLB[a,q,{ag,bg,gg,dg},v],
   Table[-coefHL[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];

(* Initial values of Hl'(z) *)

initHeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250,prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefHLPr[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefHLPrA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefHLPrB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefHLPr[a_,q_,{ag_,bg_,gg_,dg_},0]:=q/(a*gg);

coefHLPr[a_,q_,{ag_,bg_,gg_,dg_},1]:=
  (q^2-a*ag*bg*gg+q*(1+ag+bg+a*gg+(a-1)*dg))/(gg*(1+gg)*a^2);

coefHLPr[a_,q_,{ag_,bg_,gg_,dg_},n_/;(IntegerQ[n] && n>1)]:=
  coefHLPr[a,q,{ag,bg,gg,dg},n]=
  ((RecurQ[n,{a,ag,bg,gg,dg}]+q)*
   coefHLPr[a,q,{ag,bg,gg,dg},n-1]/n-RecurP[n,{ag,bg}]*
   coefHLPr[a,q,{ag,bg,gg,dg},n-2]/(n-1))/(a*(n+gg));

padeMatrixHLPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefHLPr[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixHLPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefHLPr[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefHLPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHLPrA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixHLPrA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefHLPrB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefHLPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHLPrB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixHLPrB[a,q,{ag,bg,gg,dg},v],
   Table[-coefHLPr[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];

(******************** HeunHS & HeunHSPrime (gg=!=1) ********************)

(* Local Heun functions of the 2nd kind : gg=!=1 *)

HeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHS[a,q,{ag,bg,gg,dg},z],
    realAHeunHS[a,q,{ag,bg,gg,dg},z]]/;
  (Not[intQHL[gg] && Re[gg]<2] && Precision[z]<Infinity)

HeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHSPrime[a,q,{ag,bg,gg,dg},z],
    realAHeunHSPrime[a,q,{ag,bg,gg,dg},z]]/;
  (Not[intQHL[gg] && Re[gg]<2] && Precision[z]<Infinity)

realAHeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHS[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHS[a,q,{ag,bg,gg,dg},z,-I]];

realAHeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,-I]];

compAHeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],ndsHeunHS[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]<0,ndsHeunHS[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
    Which[branchRightQ[a,z],ndsHeunHS[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]>0,ndsHeunHS[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

compAHeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]<0,ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
    Which[branchRightQ[a,z],ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]>0,ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHSPrime[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

ndsHeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHS[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHSPrime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   f[Abs[z-shift]]/.sol[[1]]];

ndsHeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHS[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHSPrime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of Hs(z): gg=!=1 *)

initHeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{eps=ag+bg-gg-dg+1},
    If[intQHL[gg] && Re[gg]>1,
      z^(1-gg)/(1-gg)*
        initReguHeunHL[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z],
      (initHeunHL[a,q,{ag,bg,gg,dg},z]-z^(1-gg)*
        initHeunHL[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z])/
        Gamma[2-gg]/(gg-1)]];

initReguHeunHL[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250-Round[gg],prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefReguHL[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefReguHLA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefReguHLB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefReguHL[a_,q_,{ag_,bg_,gg_,dg_},n_]:=0/;(n<=-Round[gg])

coefReguHL[a_,q_,{ag_,bg_,gg_,dg_},n_]:=
  coefHLNu[a,q,{ag,bg,gg,dg},n]/(Gamma[n+gg]*n!*a^n)/;(n>-Round[gg])

coefHLNu[a_,q_,{ag_,bg_,gg_,dg_},0]:=1;

coefHLNu[a_,q_,{ag_,bg_,gg_,dg_},1]:=q;

coefHLNu[a_,q_,{ag_,bg_,gg_,dg_},n_/;(IntegerQ[n] && n>1)]:=
  coefHLNu[a,q,{ag,bg,gg,dg},n]=
  (q+RecurQ[n-1,{a,ag,bg,gg,dg}])*coefHLNu[a,q,{ag,bg,gg,dg},n-1]-
  RecurP[n-1,{ag,bg}]*RecurR[n-2,{a,gg}]*coefHLNu[a,q,{ag,bg,gg,dg},n-2];

padeMatrixReguHLA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefReguHL[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixReguHLB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefReguHL[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefReguHLA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefReguHLA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixReguHLA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefReguHLB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefReguHLB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefReguHLB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixReguHLB[a,q,{ag,bg,gg,dg},v],
   Table[-coefReguHL[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];

(* Initial values of Hs'(z): gg=!=1 *)

initHeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{eps=ag+bg-gg-dg+1},
    If[intQHL[gg] && Re[gg]>1,
      z^(-gg)*initReguHeunHL[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z]+
        z^(1-gg)/(1-gg)*
        initReguHeunHLPrime[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z],
      (initHeunHLPrime[a,q,{ag,bg,gg,dg},z]/(gg-1)+
        z^(-gg)*
        initHeunHL[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z]-
        z^(1-gg)/(gg-1)*
        initHeunHLPrime[a,q-(gg-1) (a*dg+eps),{bg-gg+1,ag-gg+1,2-gg,dg},z])/
        Gamma[2-gg]]];

initReguHeunHLPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250-Round[gg],prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefReguHLPr[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefReguHLPrA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefReguHLPrB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefReguHLPr[a_,q_,{ag_,bg_,gg_,dg_},n_]:=0/;(n<=-Round[gg]-1)

coefReguHLPr[a_,q_,{ag_,bg_,gg_,dg_},n_]:=
  (n+1)*coefHLNu[a,q,{ag,bg,gg,dg},n+1]/(Gamma[gg+n+1]*(n+1)!*a^(n+1))/;
  (n>-Round[gg]-1)

padeMatrixReguHLPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefReguHLPr[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixReguHLPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefReguHLPr[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefReguHLPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefReguHLPrA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixReguHLPrA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefReguHLPrB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefReguHLPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefReguHLPrB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixReguHLPrB[a,q,{ag,bg,gg,dg},v],
   Table[-coefReguHLPr[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];

(******************** HeunHS & HeunHSPrime (gg===1) ********************)

(* Local Heun functions of the 2nd kind : gg===1 *)

HeunHS[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHS1[a,q,{ag,bg,1,dg},z],
    realAHeunHS1[a,q,{ag,bg,1,dg},z]]/;
  ((intQHL[gg] && 0.5<Re[gg]<1.5) && Precision[z]<Infinity)

HeunHSPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Abs[Im[a]]>0,
    compAHeunHS1Prime[a,q,{ag,bg,1,dg},z],
    realAHeunHS1Prime[a,q,{ag,bg,1,dg},z]]/;
  ((intQHL[gg] && 0.5<Re[gg]<1.5) && Precision[z]<Infinity)

realAHeunHS1[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHS1[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHS1[a,q,{ag,bg,gg,dg},z,-I]];

realAHeunHS1Prime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Which[Im[z]>=0,ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,I],
    Im[z]<0,ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,-I]];

compAHeunHS1[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]<0,ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
    Which[branchRightQ[a,z],ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]>0,ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS1[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

compAHeunHS1Prime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]<0,ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(a-1)/2]],
    Which[branchRightQ[a,z],ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(1+a)/2],
      Im[z]>0,ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(1-a)/2],
      branchLeftQ[a,z],ndsHeunHS1Prime[a,q,{ag,bg,gg,dg},z,(a-1)/2]]];

ndsHeunHS1[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHS1[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHS1Prime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   f[Abs[z-shift]]/.sol[[1]]];

ndsHeunHS1Prime[a_,q_,{ag_,bg_,gg_,dg_},z_,sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initHeunHS1[a,q,{ag,bg,gg,dg},shift];
    i1=initHeunHS1Prime[a,q,{ag,bg,gg,dg},shift];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)*(Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]}];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of Hs(z): gg===1 *)

initHeunHS1[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Log[z]*initHeunHL[a,q,{ag,bg,gg,dg},z]+
  initHeunHS12nd[a,q,{ag,bg,gg,dg},z];

initHeunHS12nd[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250,prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefHS12nd[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefHS12ndA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefHS12ndB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefHS12nd[a_,q_,{ag_,bg_,gg_,dg_},n_]:=
  coefHLDer[a,q,{ag,bg,gg,dg},n]-coefHL2Der[a,q,{ag,bg,gg,dg},n];

coefHLDer[a_,q_,{ag_,bg_,gg_,dg_},0]:=0;

coefHLDer[a_,q_,{ag_,bg_,gg_,dg_},1]:=-q/(a*gg^2);

coefHLDer[a_,q_,{ag_,bg_,gg_,dg_},n_/;(IntegerQ[n] && n>1)]:=
  coefHLDer[a,q,{ag,bg,gg,dg},n]=
  ((RecurQ[n-1,{a,ag,bg,gg,dg}]+q)*coefHLDer[a,q,{ag,bg,gg,dg},n-1]-
   RecurP[n-1,{ag,bg}]*coefHLDer[a,q,{ag,bg,gg,dg},n-2])/RecurR[n-1,{a,gg}]+
   (a*(n-1)/RecurR[n-1,{a,gg}]-a*n*(RecurQ[n-1,{a,ag,bg,gg,dg}]+q)/
   (RecurR[n-1,{a,gg}]^2))*coefHL[a,q,{ag,bg,gg,dg},n-1]+
   a*n*RecurP[n-1,{ag,bg}]/(RecurR[n-1,{a,gg}]^2)*
   coefHL[a,q,{ag,bg,gg,dg},n-2];

coefHL2Der[a_,q_,{ag_,bg_,gg_,dg_},0]:=0;

coefHL2Der[a_,q_,{ag_,bg_,gg_,dg_},1]:=
  (q-ag-bg-3+(4-gg) gg+(1-a) dg)/(a*(2-gg)^2);

coefHL2Der[a_,q_,{ag_,bg_,gg_,dg_},n_/;(IntegerQ[n] && n>1)]:=
  coefHL2Der[a,q,{ag,bg,gg,dg},n]=
  Module[{q1=q-(gg-1)*(a*dg+ag+bg-gg-dg+1),ag1=bg-gg+1,bg1=ag-gg+1,gg1=2-gg},
    ((RecurQ[n-1,{a,ag1,bg1,gg1,dg}]+q1)*coefHL2Der[a,q,{ag,bg,gg,dg},n-1]-
    RecurP[n-1,{ag1,bg1}]*coefHL2Der[a,q,{ag,bg,gg,dg},n-2])/RecurR[n-1,{a,gg1}]+
    ((2*(gg-n)-a (n+dg-1)-ag-bg+dg)/RecurR[n-1,{a,gg1}]+
    a*n*(RecurQ[n-1,{a,ag1,bg1,gg1,dg}]+q1)/(RecurR[n-1,{a,gg1}]^2))*
    coefHL[a,q1,{ag1,bg1,gg1,dg},n-1]-
    ((2*(gg-n+1)-ag-bg)/RecurR[n-1,{a,gg1}]+a*n*RecurP[n-1,{ag1,bg1}]/
    (RecurR[n-1,{a,gg1}]^2))*coefHL[a,q1,{ag1,bg1,gg1,dg},n-2]];

padeMatrixHS12ndA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefHS12nd[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixHS12ndB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefHS12nd[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefHS12ndA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHS12ndA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixHS12ndA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefHS12ndB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefHS12ndB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHS12ndB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixHS12ndB[a,q,{ag,bg,gg,dg},v],
   Table[-coefHS12nd[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];

(* Initial values of Hs'(z): gg===1 *)

initHeunHS1Prime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  initHeunHL[a,q,{ag,bg,gg,dg},z]/z+
   Log[z]*initHeunHLPrime[a,q,{ag,bg,gg,dg},z]+
   initHeunHS12ndPrime[a,q,{ag,bg,gg,dg},z];

initHeunHS12ndPrime[a_,q_,{ag_,bg_,gg_,dg_},z_]:=
  Module[{max=250,prec=500,ap,qp,alp,bep,gap,dep,zp,powerL},
    ap=SetPrecision[a,prec];
    qp=SetPrecision[q,prec];
    alp=SetPrecision[ag,prec];
    bep=SetPrecision[bg,prec];
    gap=SetPrecision[gg,prec];
    dep=SetPrecision[dg,prec];
    zp=SetPrecision[z,prec];
    powerL=Table[zp^j,{j,1,max,1}];
     (coefHS12ndPr[ap,qp,{alp,bep,gap,dep},0]+
     padeCoefHS12ndPrA[ap,qp,{alp,bep,gap,dep},max].powerL)/
     (1+padeCoefHS12ndPrB[ap,qp,{alp,bep,gap,dep},max].powerL)];

coefHS12ndPr[a_,q_,{ag_,bg_,gg_,dg_},n_]:=
  (n+1)*(coefHLDer[a,q,{ag,bg,gg,dg},n+1]-coefHL2Der[a,q,{ag,bg,gg,dg},n+1]);

padeMatrixHS12ndPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[If[j-k<0,0,coefHS12ndPr[a,q,{ag,bg,gg,dg},j-k]],{j,0,v,1},{k,0,v,1}];

padeMatrixHS12ndPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  Table[coefHS12ndPr[a,q,{ag,bg,gg,dg},v+j-k],{j,0,v-1,1},{k,0,v-1,1}];

padeCoefHS12ndPrA[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHS12ndPrA[a,q,{ag,bg,gg,dg},v]=
  Rest[Flatten[padeMatrixHS12ndPrA[a,q,{ag,bg,gg,dg},v].Join[{1},
   padeCoefHS12ndPrB[a,q,{ag,bg,gg,dg},v]]]];

padeCoefHS12ndPrB[a_,q_,{ag_,bg_,gg_,dg_},v_]:=
  padeCoefHS12ndPrB[a,q,{ag,bg,gg,dg},v]=
  LinearSolve[padeMatrixHS12ndPrB[a,q,{ag,bg,gg,dg},v],
   Table[-coefHS12ndPr[a,q,{ag,bg,gg,dg},v+j],{j,1,v,1}]];



(**********************************************************************************)
(*                           Generalized Heun functions                           *)
(**********************************************************************************)

(* Generalized Heun functions & it's derivative of z *)

GeneralizedHeun[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  If[Abs[Im[a]]>0,
    compAGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1}],
    realAGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1}]]/;
  (Precision[z]<Infinity)

GeneralizedHeunPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  If[Abs[Im[a]]>0,
    compAGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1}],
    realAGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1}]]/;
  (Precision[z]<Infinity)

realAGeneHeun[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},I],
    Im[z]<0,ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},-I]];

realAGeneHeunPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},I],
    Im[z]<0,ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},-I]];

compAGeneHeun[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(1+a)/2],
      Im[z]<0,
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(1-a)/2],
      branchLeftQ[a,z],
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(a-1)/2]],
    Which[branchRightQ[a,z],
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(1+a)/2],
      Im[z]>0,
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(1-a)/2],
      branchLeftQ[a,z],
        ndsGeneHeun[a,q,{ag,bg,gg,dg},z,{w0,w1},(a-1)/2]]];

compAGeneHeunPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  If[Im[a]>0,
    Which[branchRightQ[a,z],
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(1+a)/2],
      Im[z]<0,
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(1-a)/2],
      branchLeftQ[a,z],
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(a-1)/2]],
    Which[branchRightQ[a,z],
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(1+a)/2],
      Im[z]>0,
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(1-a)/2],
      branchLeftQ[a,z],
        ndsGeneHeunPrime[a,q,{ag,bg,gg,dg},z,{w0,w1},(a-1)/2]]];

ndsGeneHeun[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initGeneHeun[a,q,{ag,bg,gg,dg},shift,{w0,w1}];
    i1=initGeneHeunPrime[a,q,{ag,bg,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
      f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
      MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

ndsGeneHeunPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initGeneHeun[a,q,{ag,bg,gg,dg},shift,{w0,w1}];
    i1=initGeneHeunPrime[a,q,{ag,bg,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
      f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
      MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of the generalized Heun functions *)

initGeneHeun[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift},
    shift=If[Not[Abs[Im[a]]>0] && 0<Re[a]<=1/2,a/2,1/2];
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
      f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
      MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

initGeneHeunPrime[a_,q_,{ag_,bg_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,eps=ag+bg-gg-dg+1,shift},
    shift=If[Not[Abs[Im[a]]>0] && 0<Re[a]<=1/2,a/2,1/2];
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+
    eps/(Exp[I*theta]*x+shift-a))*Exp[-I*theta]*f'[x]+
    (ag*bg (Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1)*
    (Exp[I*theta]*x+shift-a))*f[x]==0,
      f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
      MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];



(**********************************************************************************)
(*                            Confluent Heun functions                            *)
(**********************************************************************************)

(* Confluent Heun functions & it's derivative of z *)

ConfluentHeun[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsConfHeun[q,{ag,gg,dg,eg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsConfHeun[q,{ag,gg,dg,eg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

ConfluentHeunPrime[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsConfHeunPrime[q,{ag,gg,dg,eg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsConfHeunPrime[q,{ag,gg,dg,eg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

ndsConfHeun[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initConfHeun[q,{ag,gg,dg,eg},shift,{w0,w1}];
    i1=initConfHeunPrime[q,{ag,gg,dg,eg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+eg)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

ndsConfHeunPrime[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initConfHeun[q,{ag,gg,dg,eg},shift,{w0,w1}];
    i1=initConfHeunPrime[q,{ag,gg,dg,eg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+eg)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1))*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of confluent Heun functions *)

initConfHeun[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1/2},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+eg)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1))*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

initConfHeunPrime[q_,{ag_,gg_,dg_,eg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1/2},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift-1)+eg)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift) (Exp[I*theta]*x+shift-1))*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];



(**********************************************************************************)
(*                        Doubly-Confluent Heun functions                         *)
(**********************************************************************************)

(* Doubly-Confluent Heun functions & it's derivative of z *)

DoublyConfluentHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsDoubConfHeun[q,{ag,gg,dg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsDoubConfHeun[q,{ag,gg,dg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

DoublyConfluentHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsDoubConfHeunPrime[q,{ag,gg,dg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsDoubConfHeunPrime[q,{ag,gg,dg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

ndsDoubConfHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initDoubConfHeun[q,{ag,gg,dg},shift,{w0,w1}];
    i1=initDoubConfHeunPrime[q,{ag,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift)^2+1)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)^2)*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

ndsDoubConfHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initDoubConfHeun[q,{ag,gg,dg},shift,{w0,w1}];
    i1=initDoubConfHeunPrime[q,{ag,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift)^2+1)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)^2)*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of doubly-confluent Heun functions *)

initDoubConfHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift)^2+1)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)^2)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

initDoubConfHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg/(Exp[I*theta]*x+shift)^2+1)*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    ((Exp[I*theta]*x+shift)^2)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];



(**********************************************************************************)
(*                           Biconfluent Heun functions                           *)
(**********************************************************************************)

(* Biconfluent Heun functions & it's derivative of z *)

BiConfluentHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsBiConfHeun[q,{ag,gg,dg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsBiConfHeun[q,{ag,gg,dg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

BiConfluentHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Which[Im[z]>=0,ndsBiConfHeunPrime[q,{ag,gg,dg},z,{w0,w1},(1+I)/2],
    Im[z]<0,ndsBiConfHeunPrime[q,{ag,gg,dg},z,{w0,w1},(1-I)/2]]/;
  (Precision[z]<Infinity)

ndsBiConfHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initBiConfHeun[q,{ag,gg,dg},shift,{w0,w1}];
    i1=initBiConfHeunPrime[q,{ag,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg+(Exp[I*theta]*x+shift))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    (Exp[I*theta]*x+shift)*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

ndsBiConfHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_},sh_]:=
  Module[{theta,sol,f,x,shift=N[sh],i0,i1},
    theta=If[z==shift,0,Arg[z-shift]];
    i0=initBiConfHeun[q,{ag,gg,dg},shift,{w0,w1}];
    i1=initBiConfHeunPrime[q,{ag,gg,dg},shift,{w0,w1}];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg+(Exp[I*theta]*x+shift))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    (Exp[I*theta]*x+shift)*f[x]==0,
    f[0]==i0,f'[0]==Exp[I*theta]*i1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];

(* Initial values of biconfluent Heun functions *)

initBiConfHeun[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg+(Exp[I*theta]*x+shift))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    (Exp[I*theta]*x+shift)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   f[Abs[z-shift]]/.sol[[1]]];

initBiConfHeunPrime[q_,{ag_,gg_,dg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x,shift=1},
    theta=If[z==shift,0,Arg[z-shift]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (gg/(Exp[I*theta]*x+shift)+dg+(Exp[I*theta]*x+shift))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x+shift)-q)/
    (Exp[I*theta]*x+shift)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z-shift]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z-shift]]/.sol[[1]]];



(**********************************************************************************)
(*                          Triconfluent Heun functions                           *)
(**********************************************************************************)

TriConfluentHeun[q_,{ag_,gg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x},
    theta=If[z==0,0,Arg[z]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (Exp[I*theta]*x)*(gg+(Exp[I*theta]*x))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x)-q)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z]},
    MaxStepSize->0.01];
   f[Abs[z]]/.sol[[1]]]/;
  (Precision[z]<Infinity)

TriConfluentHeunPrime[q_,{ag_,gg_},z_,{w0_,w1_}]:=
  Module[{theta,sol,f,x},
    theta=If[z==0,0,Arg[z]];
   sol=NDSolve[{Exp[-2 I*theta]*f''[x]+
    (Exp[I*theta]*x)*(gg+(Exp[I*theta]*x))*
    Exp[-I*theta]*f'[x]+(ag*(Exp[I*theta]*x)-q)*f[x]==0,
    f[0]==w0,f'[0]==Exp[I*theta]*w1},f,{x,Abs[z]},
    MaxStepSize->0.01];
   Exp[-I*theta]*f'[Abs[z]]/.sol[[1]]]/;
  (Precision[z]<Infinity)



(**********************************************************************************)

End[]

Protect[HeunHL, HeunHLPrime, HeunHS, HeunHSPrime, GeneralizedHeun, 
        GeneralizedHeunPrime, ConfluentHeun, ConfluentHeunPrime, 
        DoublyConfluentHeun, DoublyConfluentHeunPrime, BiConfluentHeun, 
        BiConfluentHeunPrime, TriConfluentHeun, TriConfluentHeunPrime]

EndPackage[]
